import numpy as np
import cv2
from pyrpoc.acquisitions.base_acquisition import Acquisition
import time
import nidaqmx
from nidaqmx.constants import AcquisitionType
from scipy.ndimage import zoom

class LocalRPOC(Acquisition):   
    def __init__(self, galvo=None, mask_data=None, treatment_parameters=None, signal_bus=None, **kwargs):
        super().__init__(**kwargs)
        self.galvo = galvo
        self.mask_data = mask_data
        self.treatment_parameters = treatment_parameters or {}
        self.signal_bus = signal_bus
        self.verified = False
        
        # Extract parameters from treatment_parameters (not acquisition_parameters)
        self.dwell_time = self.treatment_parameters.get('dwell_time', 10)
        self.extrasteps_left = self.treatment_parameters.get('extrasteps_left', 50)
        self.extrasteps_right = self.treatment_parameters.get('extrasteps_right', 50)
        self.amplitude_x = self.treatment_parameters.get('amplitude_x', 0.5)
        self.amplitude_y = self.treatment_parameters.get('amplitude_y', 0.5)
        self.offset_x = self.treatment_parameters.get('offset_x', 0.0)
        self.offset_y = self.treatment_parameters.get('offset_y', 0.0)
        self.x_pixels = self.treatment_parameters.get('x_pixels', 512)
        self.y_pixels = self.treatment_parameters.get('y_pixels', 512)
        self.offset_drift_x = self.treatment_parameters.get('offset_drift_x', 0.0)
        self.offset_drift_y = self.treatment_parameters.get('offset_drift_y', 0.0)
        self.repetitions = self.treatment_parameters.get('repetitions', 1)
        
        # Calculate treatment region from mask
        self.treatment_region = self._calculate_treatment_region()
        
    def _calculate_treatment_region(self):
        '''get a bounding box for the region'''
        if self.mask_data is None:
            return None
            
        # Find contours in the mask
        contours, _ = cv2.findContours(self.mask_data, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
        
        if not contours:
            return None
            
        # Find the bounding box that encompasses all contours
        x_min, y_min = self.x_pixels, self.y_pixels
        x_max, y_max = 0, 0
        
        for contour in contours:
            x, y, w, h = cv2.boundingRect(contour)
            x_min = min(x_min, x)
            y_min = min(y_min, y)
            x_max = max(x_max, x + w)
            y_max = max(y_max, y + h)
            
        return (x_min, y_min, x_max, y_max)
    
    def _calculate_scan_parameters(self):
        if self.treatment_region is None:
            return None
            
        x_min, y_min, x_max, y_max = self.treatment_region
        
        region_width = x_max - x_min
        region_height = y_max - y_min
        
        voltage_per_pixel_x = self.amplitude_x / self.x_pixels
        voltage_per_pixel_y = self.amplitude_y / self.y_pixels

        center_x_pixel = (x_min + x_max) / 2.0
        center_y_pixel = (y_min + y_max) / 2.0
        
        center_x_voltage = (center_x_pixel - self.x_pixels / 2.0) * voltage_per_pixel_x
        center_y_voltage = (center_y_pixel - self.y_pixels / 2.0) * voltage_per_pixel_y

        final_offset_x = self.offset_x + self.offset_drift_x
        final_offset_y = self.offset_y + self.offset_drift_y
        
        # Calculate region-specific amplitude
        region_amplitude_x = region_width * voltage_per_pixel_x
        region_amplitude_y = region_height * voltage_per_pixel_y
        
        return {
            'amplitude_x': region_amplitude_x,
            'amplitude_y': region_amplitude_y,
            'offset_x': final_offset_x + center_x_voltage,
            'offset_y': final_offset_y + center_y_voltage,
            'region_width': region_width,
            'region_height': region_height,
            'region_center_x': center_x_pixel,
            'region_center_y': center_y_pixel
        }
    
    def configure_rpoc(self, rpoc_enabled, **kwargs):
        pass
    
    def perform_acquisition(self):
        if self.signal_bus:
            self.signal_bus.console_message.emit("Starting local RPOC treatment...")
        
        # calculate scan parameters
        scan_params = self._calculate_scan_parameters()
        if scan_params is None:
            if self.signal_bus:
                self.signal_bus.console_message.emit("Error: No valid treatment region found")
            return None
        
        if self.signal_bus:
            self.signal_bus.console_message.emit(f"Treatment region: {scan_params['region_width']}x{scan_params['region_height']} pixels")
            self.signal_bus.console_message.emit(f"Center: ({scan_params['region_center_x']:.1f}, {scan_params['region_center_y']:.1f})")
        
        # Perform treatment for specified number of repetitions
        for rep in range(self.repetitions):
            if self._stop_flag and self._stop_flag():
                if self.signal_bus:
                    self.signal_bus.console_message.emit("Local RPOC treatment stopped")
                break
                
            if self.signal_bus:
                self.signal_bus.console_message.emit(f"Treatment repetition {rep + 1}/{self.repetitions}")
            
            # Perform the actual treatment scan
            self._perform_treatment_scan(scan_params)
            
            # Emit progress signal after treatment scan completes
            if self.signal_bus:
                self.signal_bus.local_rpoc_progress.emit(rep + 1)
            
            if rep < self.repetitions - 1:  # Don't wait after the last repetition
                time.sleep(0.1)  # Brief pause between repetitions
        
        if self.signal_bus:
            self.signal_bus.console_message.emit("Local RPOC treatment completed")
        
        return None  # No data to return
    
    def _perform_treatment_scan(self, scan_params):       
        try:
            rate = self.galvo.parameters.get('sample_rate', 1000000)
            dwell_time_sec = self.dwell_time / 1e6  # convert to seconds
            pixel_samples = max(1, int(dwell_time_sec * rate))
            slow_channel = self.galvo.parameters.get('slow_axis_channel', 0)
            fast_channel = self.galvo.parameters.get('fast_axis_channel', 1)
            device_name = self.galvo.parameters.get('device_name', 'Dev1')
            
            voltage_per_pixel_x = self.amplitude_x / self.x_pixels
            voltage_per_pixel_y = self.amplitude_y / self.y_pixels
            
            center_x = self.offset_x + self.offset_drift_x
            center_y = self.offset_y + self.offset_drift_y
            
            treatment_waveform, treatment_ttl = self._generate_treatment_waveform(
                center_x, center_y, voltage_per_pixel_x, voltage_per_pixel_y,
                pixel_samples, rate
            )
            
            if treatment_waveform is None:
                if self.signal_bus:
                    self.signal_bus.console_message.emit("Error: No valid treatment regions found in mask")
                return
            
                        # Send signals to galvo (similar to confocal.py)
            with nidaqmx.Task() as ao_task:                
                ao_task.ao_channels.add_ao_voltage_chan(f"{device_name}/ao{fast_channel}")
                ao_task.ao_channels.add_ao_voltage_chan(f"{device_name}/ao{slow_channel}")
                
                # Configure timing for analog output
                ao_task.timing.cfg_samp_clk_timing(
                    rate=rate,
                    sample_mode=AcquisitionType.FINITE,
                    samps_per_chan=treatment_waveform.shape[1]
                )
                
                # exported sample clock for the digital task
                # there is no AI task here so i have to explicitly wire the DO task to the AO task
                # DO tasks (apparently) do not have any implicit timing
                ao_task.export_signals.samp_clk_output_term = f"/{device_name}/PFI0"

                do_task = None
                if treatment_ttl is not None:
                    try:
                        do_task = nidaqmx.Task()

                        ttl_device = self.treatment_parameters.get('ttl_device', device_name)
                        ttl_port_line = self.treatment_parameters.get('ttl_port_line', 'port0/line0')
                        ttl_channel = f"{ttl_device}/{ttl_port_line}"
                        
                        if self.signal_bus:
                            self.signal_bus.console_message.emit(f"Debug: Configuring TTL channel: {ttl_channel}")
                        
                        do_task.do_channels.add_do_chan(ttl_channel)
                        do_task.timing.cfg_samp_clk_timing(
                            rate=rate,
                            source=f"/{device_name}/PFI0",
                            sample_mode=AcquisitionType.FINITE,
                            samps_per_chan=treatment_waveform.shape[1]
                        )
                        do_task.write(treatment_ttl.tolist(), auto_start=False)
                        
                        if self.signal_bus:
                            self.signal_bus.console_message.emit(f"TTL output configured on {ttl_channel}")
                    except Exception as e:
                        if self.signal_bus:
                            self.signal_bus.console_message.emit(f"Warning: Could not set up TTL output: {e}")
                        if do_task:
                            try:
                                do_task.close()
                            except:
                                pass
                        do_task = None

                ao_task.write(treatment_waveform, auto_start=False)

                if do_task:
                    do_task.start()
                ao_task.start()
                

                timeout = treatment_waveform.shape[1] / rate + 5
                ao_task.wait_until_done(timeout=timeout)
                if do_task:
                    do_task.wait_until_done(timeout=timeout)

                if do_task:
                    try:
                        do_task.close()
                    except:
                        pass
                
            if self.signal_bus:
                self.signal_bus.console_message.emit("Treatment scan completed successfully")
                
        except Exception as e:
            if self.signal_bus:
                self.signal_bus.console_message.emit(f"Error during treatment scan: {e}")
            time.sleep(1)
    
    def _generate_treatment_waveform(self, center_x, center_y, voltage_per_pixel_x, voltage_per_pixel_y, pixel_samples, rate):
        if self.mask_data is None:
            return None, None
        
        # Ensure mask is the right size
        mask = self.mask_data
        if mask.shape != (self.y_pixels, self.x_pixels):
            # Resize mask to match pixel dimensions
            zoom_factors = (self.y_pixels / mask.shape[0], self.x_pixels / mask.shape[1])
            mask = zoom(mask, zoom_factors, order=0).astype(bool)
        
        mask = mask > 0  # Ensure boolean mask
        
        # Find active rows (rows with any mask pixels)
        active_rows = np.any(mask, axis=1)
        if not np.any(active_rows):
            return None, None
        
        # Generate waveforms for each active row
        x_waveforms = []
        y_waveforms = []
        ttl_signals = []
        
        for row_idx in range(self.y_pixels):
            if not active_rows[row_idx]:
                continue  # Skip rows with no mask pixels
            
            # Find leftmost and rightmost mask pixels in this row
            row_mask = mask[row_idx, :]
            if not np.any(row_mask):
                continue
            
            leftmost_pixel = np.where(row_mask)[0][0]
            rightmost_pixel = np.where(row_mask)[0][-1]
            
            # Calculate voltage positions
            leftmost_voltage = center_x + (leftmost_pixel - self.x_pixels/2) * voltage_per_pixel_x
            rightmost_voltage = center_x + (rightmost_pixel - self.x_pixels/2) * voltage_per_pixel_x
            
            # Add extra steps
            extra_left_voltage = self.extrasteps_left * voltage_per_pixel_x
            extra_right_voltage = self.extrasteps_right * voltage_per_pixel_x
            
            start_voltage = leftmost_voltage - extra_left_voltage
            stop_voltage = rightmost_voltage + extra_right_voltage
            
            # Calculate number of steps for this row
            row_pixels = rightmost_pixel - leftmost_pixel + 1 + self.extrasteps_left + self.extrasteps_right
            row_samples = row_pixels * pixel_samples
            
            # Generate X waveform for this row
            x_waveform = np.linspace(start_voltage, stop_voltage, row_samples, endpoint=False)
            x_waveforms.append(x_waveform)
            
            # Generate Y position for this row
            y_position = center_y + (row_idx - self.y_pixels/2) * voltage_per_pixel_y
            y_waveform = np.full(row_samples, y_position)
            y_waveforms.append(y_waveform)
            
            # Generate TTL signal for this row
            # TTL is high only for the actual mask pixels, not extra steps
            ttl_row = np.zeros(row_samples, dtype=bool)
            
            # Calculate indices for TTL high regions
            extra_left_samples = self.extrasteps_left * pixel_samples
            mask_start_sample = extra_left_samples
            mask_end_sample = extra_left_samples + (rightmost_pixel - leftmost_pixel + 1) * pixel_samples
            
            # Set TTL high for mask pixels only
            ttl_row[mask_start_sample:mask_end_sample] = True
            
            ttl_signals.append(ttl_row)
        
        if not x_waveforms:
            return None, None
        
        # Concatenate all waveforms
        x_composite = np.concatenate(x_waveforms)
        y_composite = np.concatenate(y_waveforms)
        ttl_composite = np.concatenate(ttl_signals)
        
        # Create final waveform
        treatment_waveform = np.vstack([x_composite, y_composite])
        
        if self.signal_bus:
            total_samples = treatment_waveform.shape[1]
            total_time = total_samples / rate
            self.signal_bus.console_message.emit(f"Treatment waveform: {len(x_waveforms)} active rows, {total_samples} samples, {total_time:.3f}s")
        
        return treatment_waveform, ttl_composite
    
    def save_data(self, data):
        """
        Local RPOC doesn't save data - this is a placeholder
        """
        pass



"""
Merge Time Tag Stream files recorded separately from two Time Taggers.

This example records two time tag stream files, one for channels 1 and 2 and another for channels 3 and 4.
Later these separate stream files are combined into one to calculate the correlation between channels 1 and 3.
The file merger is combining the streams. The user can specify a constant time offset for each stream
as well as the channel number offset.

"""

from matplotlib import pyplot as plt

from Swabian.TimeTagger import (createTimeTagger, createTimeTaggerVirtual, mergeStreamFiles,
                                FileWriter, SynchronizedMeasurements, Correlation)


MAKE_TESTDATA = True
MERGE_DUMPS = True
REPLAY_MERGE = True

# Create test data
if MAKE_TESTDATA:
    with createTimeTagger() as t:
        t.setTestSignal([1, 2, 3, 4], True)
        with SynchronizedMeasurements(t) as sm:
            fw12 = FileWriter(sm.getTagger(), 'dump12.ttbin', [1, 2])
            fw34 = FileWriter(sm.getTagger(), 'dump34.ttbin', [3, 4])
            sm.startFor(5e12)
            sm.waitUntilFinished()


if MERGE_DUMPS:
    # Merges multiple *.ttbin files into one.
    # You can specify channel number offsets to separate channels if dumps were recorded using the same channel numbers
    # You can specify a constant time offset for each ttbin file.
    mergeStreamFiles(
        output_filename='dump1234_merged.ttbin',            # Filename of the output ttbin file
        input_filenames=['dump12.ttbin', 'dump34.ttbin'],   # List of dump files that will be merged
        channel_offsets=[0, 0],     # Channel number offset for each ttbin file. Useful when dumps have the same channel numbers.
        time_offsets=[0, 1000],     # Time offset for each ttbin file in picoseconds.
        overlap_only=True,          # If True, then merge only the regions where the time is overlapping.
    )


if REPLAY_MERGE:
    # Uses merged file and calculates correlation for the channels 1&3
    # that were originally in different dump files, and also for channels 1&2.
    with createTimeTaggerVirtual('dump1234_merged.ttbin') as ttv:
        cor12 = Correlation(ttv, 1, 2, binwidth=10, n_bins=3000)
        cor13 = Correlation(ttv, 1, 3, binwidth=10, n_bins=3000)

        ttv.run()
        ttv.waitUntilFinished()

    plt.plot(cor12.getIndex(), cor12.getData(), label='1 vs 2')
    plt.plot(cor13.getIndex(), cor13.getData(), label='1 vs 3')
    plt.xlabel('Time (ps)')
    plt.ylabel('Counts/bin')
    plt.legend()

    plt.show()



from __future__ import annotations

from PyQt6.QtCore import QObject

# Ensure decorators run and registries are populated.
import pyrpoc.displays  # noqa: F401
import pyrpoc.instruments  # noqa: F401
import pyrpoc.modalities  # noqa: F401
import pyrpoc.optocontrols  # noqa: F401

from pyrpoc.domain.app_state import AppState
from pyrpoc.gui.main_gui import MainGUI
from pyrpoc.gui.styles.theme_manager import ThemeController
from pyrpoc.persistence.session_repository import SessionRepository
from .display_service import DisplayService
from .instrument_service import InstrumentService
from .modality_service import ModalityService
from .opto_control_service import OptoControlService
from .session_coordinator import SessionCoordinator


class AppController(QObject):
    def __init__(self, theme_controller: ThemeController, parent=None):
        super().__init__(parent)
        self.app_state = AppState()

        self.instrument_service = InstrumentService(self.app_state, self)
        self.modality_service = ModalityService(self.instrument_service, self.app_state, self)
        self.display_service = DisplayService(self.app_state, self)
        self.opto_control_service = OptoControlService(self.app_state, self)

        # Acquisition-to-display flow:
        # - modality acquires one frame and emits `data_ready`
        # - this connection routes that payload into display fanout rendering.
        self.modality_service.data_ready.connect(self.display_service.push_data)
        self.instrument_service.inventory_changed.connect(
            lambda *_: self.modality_service.validate_required_instruments()
        )

        self.main_window = MainGUI(
            instrument_service=self.instrument_service,
            modality_service=self.modality_service,
            display_service=self.display_service,
            opto_control_service=self.opto_control_service,
            theme_controller=theme_controller,
        )
        self.session_coordinator = SessionCoordinator(
            app_state=self.app_state,
            repository=SessionRepository(),
            theme_controller=theme_controller,
            instrument_service=self.instrument_service,
            modality_service=self.modality_service,
            display_service=self.display_service,
            opto_control_service=self.opto_control_service,
            main_window=self.main_window,
            parent=self,
        )
        self.main_window.menubar.new_requested.connect(self.session_coordinator.reset_session)
        self.main_window.menubar.open_requested.connect(self.session_coordinator.restore_on_startup)
        self.main_window.menubar.save_requested.connect(self.session_coordinator.save_now)
        self.main_window.menubar.save_as_requested.connect(self.session_coordinator.save_now)
        self.main_window.closing.connect(self.session_coordinator.save_now)

    def show(self) -> None:
        self.main_window.show()
        self.session_coordinator.restore_on_startup()



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass
class ObjectStore(Generic[T]):
    _items: list[T]

    def __init__(self, items: list[T]):
        self._items = items

    def add(self, item: T) -> T:
        self._items.append(item)
        return item

    def remove(self, item: T) -> None:
        if item in self._items:
            self._items.remove(item)

    def list(self) -> list[T]:
        return list(self._items)

    def contains(self, item: T) -> bool:
        return item in self._items

    def clear(self) -> list[T]:
        removed = list(self._items)
        self._items.clear()
        return removed

    def index_of(self, item: T) -> int:
        return self._items.index(item)



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any

from pyrpoc.displays.base_display import BaseDisplay
from pyrpoc.instruments.base_instrument import BaseInstrument
from pyrpoc.modalities.base_modality import BaseModality
from pyrpoc.optocontrols.base_optocontrol import BaseOptoControl


@dataclass
class ParameterValue:
    label: str
    value: Any


@dataclass
class ModalityState:
    selected_key: str | None = None
    selected_class: type[BaseModality] | None = None
    instance: BaseModality | None = None
    configured_params: list[ParameterValue] = field(default_factory=list)
    running: bool = False
    last_error: str | None = None


@dataclass
class GuiLayoutState:
    ads_state_base64: str | None = None
    dock_visibility: dict[str, bool] = field(default_factory=dict)
    expanded_opto_index: int | None = None


@dataclass
class AppState:
    # Instance-first inventory model:
    # - each list entry is the concrete runtime object
    # - service rows still expose this object via row["state"] for UI compatibility
    instruments: list[BaseInstrument] = field(default_factory=list)
    optocontrols: list[BaseOptoControl] = field(default_factory=list)
    displays: list[BaseDisplay] = field(default_factory=list)
    modality: ModalityState = field(default_factory=ModalityState)
    gui_layout: GuiLayoutState = field(default_factory=GuiLayoutState)



from __future__ import annotations

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QAbstractItemView, QWidget

from pyrpoc.displays.base_display import BaseDisplay
from pyrpoc.gui.main_widgets.display_mgr.handlers import (
    on_add_clicked,
    on_attach_clicked,
    on_detach_clicked,
    on_display_name_edited,
    on_display_error,
    on_modality_selected,
    on_remove_clicked,
    refresh_available,
    refresh_instances,
    show_error,
)
from pyrpoc.gui.main_widgets.display_mgr.state import DisplayManagerState
from pyrpoc.gui.main_widgets.display_mgr.ui import build_display_manager_ui
from pyrpoc.services.display_service import DisplayService
from pyrpoc.services.modality_service import ModalityService


class DisplayManagerWidget(QWidget):
    def __init__(
        self,
        display_service: DisplayService,
        modality_service: ModalityService,
        parent: QWidget | None = None,
    ):
        super().__init__(parent)
        self.display_service = display_service
        self.modality_service = modality_service
        self.state = DisplayManagerState()
        self.ui = build_display_manager_ui(self)

        # Compatibility aliases for existing callers.
        self.display_combo = self.ui.display_combo
        self.add_btn = self.ui.add_btn
        self.instances_list = self.ui.instances_list
        self.instances_list.setEditTriggers(
            QAbstractItemView.EditTrigger.AllEditTriggers
        )
        self.instances_list.itemDoubleClicked.connect(self._on_display_name_edit_request)
        self.attach_btn = self.ui.attach_btn
        self.detach_btn = self.ui.detach_btn
        self.remove_btn = self.ui.remove_btn
        self.status_label = self.ui.status_label
        self.name_input = self.ui.name_input

        self._wire_signals()
        self._refresh_available()
        self._refresh_instances()

    def _wire_signals(self) -> None:
        self.add_btn.clicked.connect(self._on_add_clicked)
        self.attach_btn.clicked.connect(self._on_attach_clicked)
        self.detach_btn.clicked.connect(self._on_detach_clicked)
        self.remove_btn.clicked.connect(self._on_remove_clicked)

        self.display_service.display_added.connect(lambda _state: self._refresh_instances())
        self.display_service.display_removed.connect(lambda _state: self._refresh_instances())
        self.display_service.display_changed.connect(lambda _state: self._refresh_instances())
        self.display_service.display_error.connect(self._on_display_error)
        self.instances_list.itemChanged.connect(self._on_display_name_edited)
        self.modality_service.modality_selected.connect(self._on_modality_selected)

    def _refresh_available(self) -> None:
        refresh_available(self)

    def _refresh_instances(self) -> None:
        refresh_instances(self)

    def _on_add_clicked(self) -> None:
        on_add_clicked(self)

    def _on_attach_clicked(self) -> None:
        on_attach_clicked(self)

    def _on_detach_clicked(self) -> None:
        on_detach_clicked(self)

    def _on_remove_clicked(self) -> None:
        on_remove_clicked(self)

    def _on_display_error(self, state: object, message: str) -> None:
        on_display_error(self, state, message)

    def _on_display_name_edit_request(self, item) -> None:
        if item is None:
            return
        self.instances_list.editItem(item)

    def _on_modality_selected(self, key: str) -> None:
        on_modality_selected(self, key)

    def _selected_display(self) -> BaseDisplay | None:
        item = self.instances_list.currentItem()
        if item is None:
            return None
        value = item.data(Qt.ItemDataRole.UserRole)
        if not isinstance(value, int):
            return None
        display = self.display_service.get_display_by_id(value)
        return display

    def _selected_display_key(self) -> str:
        data = self.display_combo.currentData()
        if isinstance(data, str):
            return data
        return self.display_combo.currentText().strip()

    def _show_error(self, message: str) -> None:
        show_error(self, message)

    def _on_display_name_edited(self, item) -> None:
        on_display_name_edited(self, item)



import time
from pathlib import Path

import nidaqmx
import numpy as np
import tifffile
from nidaqmx.constants import AcquisitionType, LineGrouping

from .base_acquisition import Acquisition


class FLIM(Acquisition):
    def __init__(self, galvo=None, data_inputs=None, num_frames=1, signal_bus=None, acquisition_parameters=None, **kwargs):
        super().__init__(**kwargs)
        self.galvo = galvo
        self.data_inputs = data_inputs or []
        self.num_frames = num_frames
        self.signal_bus = signal_bus
        self.verified = False
        self.acquisition_parameters = acquisition_parameters or {}

    def configure_rpoc(self, rpoc_enabled, **kwargs):
        # FLIM no longer carries RPOC-specific acquisition behavior.
        return None

    def perform_acquisition(self):
        self.save_metadata()

        ai_channels = []
        for data_input in self.data_inputs:
            channels = data_input.parameters.get('input_channels', [])
            device = data_input.parameters.get('device_name', 'Dev1')
            for ch in channels:
                ai_channels.append(f"{device}/ai{ch}")

        if not ai_channels:
            ai_channels = ["Dev1/ai0"]

        all_frames = []
        for frame_idx in range(self.num_frames):
            if self._stop_flag and self._stop_flag():
                break

            frame_data = self.collect_data(self.galvo, ai_channels)

            if self.signal_bus:
                self.signal_bus.data_signal.emit(frame_data, frame_idx, self.num_frames, False)
            all_frames.append(frame_data)

        if all_frames:
            final_data = np.stack(all_frames)
            if self.signal_bus:
                self.signal_bus.data_signal.emit(final_data, len(all_frames) - 1, self.num_frames, True)
            self.save_data(final_data)
            return final_data
        return None

    def generate_simulated_confocal(self):
        x_pixels = self.acquisition_parameters.get('x_pixels', 512)
        y_pixels = self.acquisition_parameters.get('y_pixels', 512)

        num_channels = 1
        if self.data_inputs:
            total_channels = 0
            for data_input in self.data_inputs:
                channels = data_input.parameters.get('input_channels', [])
                total_channels += len(channels)
            num_channels = max(1, total_channels)

        frame = np.zeros((num_channels, y_pixels, x_pixels))

        for ch in range(num_channels):
            num_circles = np.random.randint(2, 6)
            for _ in range(num_circles):
                center_x = np.random.randint(50, x_pixels - 50)
                center_y = np.random.randint(50, y_pixels - 50)
                radius = np.random.randint(10, 30)
                intensity = np.random.uniform(0.3, 1.0)

                y, x = np.ogrid[:y_pixels, :x_pixels]
                mask = (x - center_x) ** 2 + (y - center_y) ** 2 <= radius ** 2
                frame[ch, mask] += intensity

            frame[ch] += np.random.normal(0, 0.1, frame[ch].shape)
            frame[ch] = np.clip(frame[ch], 0, 1)

        time.sleep(1)
        return frame

    def _build_pixel_start_signal(self, total_y, total_x, pixel_samples, extra_left, numsteps_x):
        if not self.acquisition_parameters.get('pixel_start_enabled', False):
            return None

        pulse_samples = int(self.acquisition_parameters.get('pixel_start_pulse_samples', 1))
        pulse_samples = max(1, min(pixel_samples, pulse_samples))

        ttl = np.zeros((total_y, total_x, pixel_samples), dtype=bool)
        active_start = extra_left
        active_stop = extra_left + numsteps_x
        ttl[:, active_start:active_stop, :pulse_samples] = True
        return ttl.ravel()

    def _get_pixel_start_line(self):
        if not self.acquisition_parameters.get('pixel_start_enabled', False):
            return None

        device_name = self.acquisition_parameters.get(
            'pixel_start_device_name',
            self.galvo.parameters.get('device_name', 'Dev1'),
        )
        port = self.acquisition_parameters.get('pixel_start_port', 'port0')
        line = self.acquisition_parameters.get('pixel_start_line', 0)
        return f"{device_name}/{port}/line{line}"

    def collect_data(self, galvo, ai_channels):
        try:
            rate = galvo.parameters.get('sample_rate', 1000000)
            dwell_time = self.acquisition_parameters.get('dwell_time', 10)
            dwell_time_sec = dwell_time / 1e6
            extra_left = self.acquisition_parameters.get('extrasteps_left', 50)
            extra_right = self.acquisition_parameters.get('extrasteps_right', 50)
            numsteps_x = self.acquisition_parameters.get('x_pixels', 512)
            numsteps_y = self.acquisition_parameters.get('y_pixels', 512)
            slow_channel = galvo.parameters.get('slow_axis_channel', 0)
            fast_channel = galvo.parameters.get('fast_axis_channel', 1)
            device_name = galvo.parameters.get('device_name', 'Dev1')

            pixel_samples = max(1, int(dwell_time_sec * rate))
            total_x = numsteps_x + extra_left + extra_right
            total_y = numsteps_y
            total_samples = total_x * total_y * pixel_samples

            waveform = galvo.generate_raster_waveform(self.acquisition_parameters)
            pixel_start_signal = self._build_pixel_start_signal(
                total_y=total_y,
                total_x=total_x,
                pixel_samples=pixel_samples,
                extra_left=extra_left,
                numsteps_x=numsteps_x,
            )
            pixel_start_line = self._get_pixel_start_line()

            with nidaqmx.Task() as ao_task, nidaqmx.Task() as ai_task:
                ao_task.ao_channels.add_ao_voltage_chan(f"{device_name}/ao{fast_channel}")
                ao_task.ao_channels.add_ao_voltage_chan(f"{device_name}/ao{slow_channel}")
                for ch in ai_channels:
                    ai_task.ai_channels.add_ai_voltage_chan(ch)

                ao_task.timing.cfg_samp_clk_timing(
                    rate=rate,
                    sample_mode=AcquisitionType.FINITE,
                    samps_per_chan=total_samples,
                )
                ai_task.timing.cfg_samp_clk_timing(
                    rate=rate,
                    source=f"/{device_name}/ao/SampleClock",
                    sample_mode=AcquisitionType.FINITE,
                    samps_per_chan=total_samples,
                )

                pixel_start_task = None
                if pixel_start_signal is not None and pixel_start_line is not None:
                    pixel_start_task = nidaqmx.Task()
                    pixel_start_task.do_channels.add_do_chan(
                        pixel_start_line,
                        line_grouping=LineGrouping.CHAN_PER_LINE,
                    )
                    pixel_start_task.timing.cfg_samp_clk_timing(
                        rate=rate,
                        source=f"/{device_name}/ao/SampleClock",
                        sample_mode=AcquisitionType.FINITE,
                        samps_per_chan=total_samples,
                    )
                    pixel_start_task.write(pixel_start_signal.tolist(), auto_start=False)

                ao_task.write(waveform, auto_start=False)
                ai_task.start()
                if pixel_start_task:
                    pixel_start_task.start()
                ao_task.start()

                timeout = total_samples / rate + 5
                ao_task.wait_until_done(timeout=timeout)
                ai_task.wait_until_done(timeout=timeout)
                if pixel_start_task:
                    pixel_start_task.wait_until_done(timeout=timeout)
                    pixel_start_task.close()

                acq_data = np.array(ai_task.read(number_of_samples_per_channel=total_samples))

                results = []
                for i in range(len(ai_channels)):
                    channel_data = acq_data if len(ai_channels) == 1 else acq_data[i]
                    reshaped = channel_data.reshape(total_y, total_x, pixel_samples)
                    pixel_values = np.mean(reshaped, axis=2)
                    cropped = pixel_values[:, extra_left:extra_left + numsteps_x]
                    results.append(cropped)

                if len(results) == 1:
                    return results[0]
                return np.stack(results)

        except Exception as e:
            if self.signal_bus:
                self.signal_bus.console_message.emit(f'Error in DAQ acquisition: {e}')
            return self.generate_simulated_confocal()

    def save_data(self, data):
        if not self.save_enabled or not self.save_path:
            return

        try:
            save_dir = Path(self.save_path).parent
            if not save_dir.is_dir():
                save_dir.mkdir(parents=True, exist_ok=True)

            if len(data.shape) == 4:
                num_frames, num_channels, height, width = data.shape
            elif len(data.shape) == 3:
                num_frames, height, width = data.shape
                num_channels = 1
                data = data.reshape(num_frames, 1, height, width)
            else:
                raise ValueError(f"Unexpected data shape: {data.shape}")

            channel_names = self._get_channel_names()

            for ch in range(num_channels):
                channel_data = data[:, ch, :, :]
                channel_name = channel_names.get(ch, f"ch{ch:02d}")
                filename = f"{Path(self.save_path).stem}_{channel_name}.tiff"
                filepath = save_dir / filename
                tifffile.imwrite(filepath, channel_data)

        except Exception as e:
            if self.signal_bus:
                self.signal_bus.console_message.emit(f"Error saving FLIM data: {e}")

    def _get_channel_names(self):
        channel_names = {}
        channel_idx = 0

        if hasattr(self, 'data_inputs') and self.data_inputs:
            for data_input in self.data_inputs:
                if hasattr(data_input, 'parameters'):
                    input_channels = data_input.parameters.get('input_channels', [])
                    channel_names_param = data_input.parameters.get('channel_names', {})

                    for ch in input_channels:
                        channel_name = channel_names_param.get(str(ch), f"ch{ch}")
                        channel_names[channel_idx] = channel_name
                        channel_idx += 1

        if not channel_names:
            for ch in range(channel_idx):
                channel_names[ch] = f"ch{ch:02d}"

        return channel_names



from __future__ import annotations

from typing import Any, Callable

from PyQt6.QtWidgets import QFormLayout, QGroupBox, QLabel, QPushButton, QVBoxLayout, QWidget

from pyrpoc.backend_utils.contracts import Action
from pyrpoc.backend_utils.parameter_utils import BaseParameter
from pyrpoc.gui.main_widgets.instrument_mgr.state import InstrumentManagerState
from pyrpoc.gui.main_widgets.instrument_mgr.ui import InstrumentManagerUI


def build_config_form(
    ui: InstrumentManagerUI,
    state: InstrumentManagerState,
    parameter_groups: dict[str, list[BaseParameter]],
) -> None:
    state.config_widgets.clear()
    while ui.config_form.rowCount() > 0:
        ui.config_form.removeRow(0)

    for group_name, parameters in parameter_groups.items():
        if group_name:
            header = QLabel(f"[{group_name.capitalize()}]", ui.config_box)
            header.setStyleSheet("font-weight: bold;")
            ui.config_form.addRow(header)
        for param in parameters:
            editor = param.create_widget(ui.config_box)
            state.config_widgets[param.label] = editor
            state.config_defs[param.label] = param
            ui.config_form.addRow(param.display_label, editor)


def build_actions_area(
    ui: InstrumentManagerUI,
    state: InstrumentManagerState,
    actions: list[Action],
    run_action: Callable[[str], None],
) -> None:
    state.action_widgets.clear()
    state.actions_by_label.clear()
    state.actions_by_label.update({action.label: action for action in actions})

    while ui.actions_layout.count():
        item = ui.actions_layout.takeAt(0)
        widget = item.widget()
        if widget is not None:
            widget.deleteLater()

    for action in actions:
        action_box = QGroupBox(action.label, ui.actions_box)
        box_layout = QVBoxLayout(action_box)

        if action.tooltip:
            action_box.setToolTip(action.tooltip)

        form = QFormLayout()
        action_param_widgets: dict[str, tuple[BaseParameter, QWidget]] = {}
        for param in action.parameters:
            editor = param.create_widget(action_box)
            action_param_widgets[param.label] = (param, editor)
            form.addRow(param.display_label, editor)

        box_layout.addLayout(form)

        run_btn = QPushButton("Run", action_box)
        run_btn.clicked.connect(lambda checked, action_label=action.label: run_action(action_label))
        box_layout.addWidget(run_btn)

        state.action_widgets[action.label] = action_param_widgets
        ui.actions_layout.addWidget(action_box)

    ui.actions_layout.addStretch(1)


def collect_values(widget_map: dict[str, tuple[BaseParameter, QWidget]]) -> dict[str, Any]:
    values: dict[str, Any] = {}
    for label, (param, widget) in widget_map.items():
        values[label] = param.get_value(widget)
    return values



from PyQt6.QtWidgets import QDialog, QVBoxLayout, QHBoxLayout, QLabel, QPushButton, QGroupBox, QFormLayout, QDoubleSpinBox, QSpinBox, QProgressBar, QComboBox
from PyQt6.QtCore import pyqtSignal, QTimer
from superqt import QSearchableComboBox

class LocalRPOCProgressDialog(QDialog):
    """Progress dialog for local RPOC treatment"""
    
    def __init__(self, total_repetitions, parent=None):
        super().__init__(parent)
        self.setWindowTitle("Local RPOC Treatment Progress")
        self.setModal(True)
        self.setFixedSize(400, 150)
        
        layout = QVBoxLayout()
        
        # Status label
        self.status_label = QLabel("Preparing treatment...")
        layout.addWidget(self.status_label)
        
        # Progress bar
        self.progress_bar = QProgressBar()
        self.progress_bar.setRange(0, total_repetitions)
        self.progress_bar.setValue(0)
        layout.addWidget(self.progress_bar)
        
        # Repetition counter
        self.repetition_label = QLabel(f"Repetition: 0/{total_repetitions}")
        layout.addWidget(self.repetition_label)
        
        # Cancel button
        self.cancel_btn = QPushButton("Cancel")
        self.cancel_btn.clicked.connect(self.reject)
        layout.addWidget(self.cancel_btn)
        
        self.setLayout(layout)
        
        self.total_repetitions = total_repetitions
        self.current_repetition = 0
    
    def update_progress(self, repetition_number):
        """Update progress bar and labels"""
        self.current_repetition = repetition_number
        self.progress_bar.setValue(repetition_number)
        self.repetition_label.setText(f"Repetition: {repetition_number}/{self.total_repetitions}")
        self.status_label.setText(f"Treatment in progress... (Repetition {repetition_number})")
    
    def set_completed(self):
        """Mark treatment as completed"""
        self.status_label.setText("Treatment completed!")
        self.progress_bar.setValue(self.total_repetitions)
        self.repetition_label.setText(f"Repetition: {self.total_repetitions}/{self.total_repetitions}")
        self.cancel_btn.setText("Close")
    
    def set_stopped(self):
        """Mark treatment as stopped"""
        self.status_label.setText("Treatment stopped.")
        self.cancel_btn.setText("Close")

class LocalRPOCDialog(QDialog):
    """Dialog for configuring local RPOC parameters"""
    
    def __init__(self, app_state, signals, parent=None):
        super().__init__(parent)
        self.app_state = app_state
        self.signals = signals
        self.setWindowTitle("Local RPOC Configuration")
        self.setModal(True)
        self.resize(500, 600)
        
        layout = QVBoxLayout()
        
        # Acquisition parameters group
        self.add_acquisition_parameters(layout)
        
        # Drift offset parameters group
        self.add_drift_parameters(layout)
        
        # Repetitions parameter
        self.add_repetitions_parameter(layout)
        
        # TTL channel selection
        self.add_ttl_channel_selection(layout)
        
        # Buttons
        button_layout = QHBoxLayout()
        start_btn = QPushButton("Start Local RPOC")
        start_btn.clicked.connect(self.start_local_rpoc)
        cancel_btn = QPushButton("Cancel")
        cancel_btn.clicked.connect(self.reject)
        
        button_layout.addWidget(start_btn)
        button_layout.addWidget(cancel_btn)
        layout.addLayout(button_layout)
        
        self.setLayout(layout)
    
    def add_acquisition_parameters(self, layout):
        """Add acquisition parameters from app_state"""
        acq_group = QGroupBox("Acquisition Parameters")
        acq_layout = QFormLayout()
        
        # Dwell time
        self.dwell_time_spin = QDoubleSpinBox()
        self.dwell_time_spin.setRange(1, 1000)
        self.dwell_time_spin.setValue(self.app_state.acquisition_parameters.get('dwell_time', 10))
        self.dwell_time_spin.setSuffix(" μs")
        self.dwell_time_spin.setDecimals(0)
        self.dwell_time_spin.setSingleStep(1)
        acq_layout.addRow("Dwell Time:", self.dwell_time_spin)
        
        # Extra steps
        self.extrasteps_left_spin = QSpinBox()
        self.extrasteps_left_spin.setRange(0, 10000)
        self.extrasteps_left_spin.setValue(self.app_state.acquisition_parameters.get('extrasteps_left', 50))
        acq_layout.addRow("Extra Steps Left:", self.extrasteps_left_spin)
        
        self.extrasteps_right_spin = QSpinBox()
        self.extrasteps_right_spin.setRange(0, 10000)
        self.extrasteps_right_spin.setValue(self.app_state.acquisition_parameters.get('extrasteps_right', 50))
        acq_layout.addRow("Extra Steps Right:", self.extrasteps_right_spin)
        
        # Amplitude
        self.amplitude_x_spin = QDoubleSpinBox()
        self.amplitude_x_spin.setRange(0.0, 10.0)
        self.amplitude_x_spin.setDecimals(1)
        self.amplitude_x_spin.setSingleStep(0.1)
        self.amplitude_x_spin.setValue(self.app_state.acquisition_parameters.get('amplitude_x', 0.5))
        self.amplitude_x_spin.setSuffix(" V")
        acq_layout.addRow("Amplitude X:", self.amplitude_x_spin)
        
        self.amplitude_y_spin = QDoubleSpinBox()
        self.amplitude_y_spin.setRange(0.0, 10.0)
        self.amplitude_y_spin.setDecimals(1)
        self.amplitude_y_spin.setSingleStep(0.1)
        self.amplitude_y_spin.setValue(self.app_state.acquisition_parameters.get('amplitude_y', 0.5))
        self.amplitude_y_spin.setSuffix(" V")
        acq_layout.addRow("Amplitude Y:", self.amplitude_y_spin)
        
        # Offset
        self.offset_x_spin = QDoubleSpinBox()
        self.offset_x_spin.setRange(-10.0, 10.0)
        self.offset_x_spin.setDecimals(1)
        self.offset_x_spin.setSingleStep(0.1)
        self.offset_x_spin.setValue(self.app_state.acquisition_parameters.get('offset_x', 0.0))
        self.offset_x_spin.setSuffix(" V")
        acq_layout.addRow("Offset X:", self.offset_x_spin)
        
        self.offset_y_spin = QDoubleSpinBox()
        self.offset_y_spin.setRange(-10.0, 10.0)
        self.offset_y_spin.setDecimals(1)
        self.offset_y_spin.setSingleStep(0.1)
        self.offset_y_spin.setValue(self.app_state.acquisition_parameters.get('offset_y', 0.0))
        self.offset_y_spin.setSuffix(" V")
        acq_layout.addRow("Offset Y:", self.offset_y_spin)
        
        # Pixels
        self.x_pixels_spin = QSpinBox()
        self.x_pixels_spin.setRange(64, 4096)
        self.x_pixels_spin.setValue(self.app_state.acquisition_parameters.get('x_pixels', 512))
        acq_layout.addRow("X Pixels:", self.x_pixels_spin)
        
        self.y_pixels_spin = QSpinBox()
        self.y_pixels_spin.setRange(64, 4096)
        self.y_pixels_spin.setValue(self.app_state.acquisition_parameters.get('y_pixels', 512))
        acq_layout.addRow("Y Pixels:", self.y_pixels_spin)
        
        acq_group.setLayout(acq_layout)
        layout.addWidget(acq_group)
    
    def add_drift_parameters(self, layout):
        """Add drift offset parameters"""
        drift_group = QGroupBox("Drift Offset Parameters")
        drift_layout = QFormLayout()
        
        self.offset_drift_x_spin = QDoubleSpinBox()
        self.offset_drift_x_spin.setRange(-10.0, 10.0)
        self.offset_drift_x_spin.setDecimals(3)
        self.offset_drift_x_spin.setSingleStep(0.001)
        self.offset_drift_x_spin.setValue(0.0)
        self.offset_drift_x_spin.setSuffix(" V")
        drift_layout.addRow("Offset Drift X:", self.offset_drift_x_spin)
        
        self.offset_drift_y_spin = QDoubleSpinBox()
        self.offset_drift_y_spin.setRange(-10.0, 10.0)
        self.offset_drift_y_spin.setDecimals(3)
        self.offset_drift_y_spin.setSingleStep(0.001)
        self.offset_drift_y_spin.setValue(0.0)
        self.offset_drift_y_spin.setSuffix(" V")
        drift_layout.addRow("Offset Drift Y:", self.offset_drift_y_spin)
        
        drift_group.setLayout(drift_layout)
        layout.addWidget(drift_group)
    
    def add_repetitions_parameter(self, layout):
        """Add number of repetitions parameter"""
        reps_layout = QHBoxLayout()
        reps_layout.addWidget(QLabel("Number of Repetitions:"))
        
        self.repetitions_spin = QSpinBox()
        self.repetitions_spin.setRange(1, 1000)
        self.repetitions_spin.setValue(1)
        reps_layout.addWidget(self.repetitions_spin)
        
        layout.addLayout(reps_layout)
    
    def add_ttl_channel_selection(self, layout):
        """Add TTL channel selection"""
        ttl_group = QGroupBox("TTL Channel Selection")
        ttl_layout = QFormLayout()
        
        # Device selection
        self.ttl_device_combo = QSearchableComboBox()
        self.ttl_device_combo.addItems(['Dev1', 'Dev2', 'Dev3', 'Dev4'])
        
        # Port/Line selection
        self.ttl_port_line_combo = QSearchableComboBox()
        self.ttl_port_line_combo.addItems([
            'port0/line0', 'port0/line1', 'port0/line2', 'port0/line3', 'port0/line4', 'port0/line5', 'port0/line6', 'port0/line7',
            'port0/line8', 'port0/line9', 'port0/line10', 'port0/line11', 'port0/line12', 'port0/line13', 'port0/line14', 'port0/line15',
            'port1/line0', 'port1/line1', 'port1/line2', 'port1/line3', 'port1/line4', 'port1/line5', 'port1/line6', 'port1/line7',
            'port1/line8', 'port1/line9', 'port1/line10', 'port1/line11', 'port1/line12', 'port1/line13', 'port1/line14', 'port1/line15'
        ])
        
        # Set default values from the parent widget (RPOC channel)
        if hasattr(self.parent(), 'channel_id') and hasattr(self.parent(), 'app_state'):
            channel_id = self.parent().channel_id
            if hasattr(self.parent().app_state, 'rpoc_mask_channels') and channel_id in self.parent().app_state.rpoc_mask_channels:
                channel_data = self.parent().app_state.rpoc_mask_channels[channel_id]
                default_device = channel_data.get('device', 'Dev1')
                default_port_line = channel_data.get('port_line', 'port0/line0')
                
                self.ttl_device_combo.setCurrentText(default_device)
                self.ttl_port_line_combo.setCurrentText(default_port_line)
        
        ttl_layout.addRow("Device:", self.ttl_device_combo)
        ttl_layout.addRow("Port/Line:", self.ttl_port_line_combo)
        
        ttl_group.setLayout(ttl_layout)
        layout.addWidget(ttl_group)
    
    def get_parameters(self):
        """Get all parameters from the dialog"""
        return {
            'dwell_time': self.dwell_time_spin.value(),
            'extrasteps_left': self.extrasteps_left_spin.value(),
            'extrasteps_right': self.extrasteps_right_spin.value(),
            'amplitude_x': self.amplitude_x_spin.value(),
            'amplitude_y': self.amplitude_y_spin.value(),
            'offset_x': self.offset_x_spin.value(),
            'offset_y': self.offset_y_spin.value(),
            'x_pixels': self.x_pixels_spin.value(),
            'y_pixels': self.y_pixels_spin.value(),
            'offset_drift_x': self.offset_drift_x_spin.value(),
            'offset_drift_y': self.offset_drift_y_spin.value(),
            'repetitions': self.repetitions_spin.value(),
            'ttl_device': self.ttl_device_combo.currentText(),
            'ttl_port_line': self.ttl_port_line_combo.currentText()
        }
    
    def start_local_rpoc(self):
        """Start the local RPOC treatment"""
        parameters = self.get_parameters()
        self.signals.local_rpoc_started.emit(parameters)
        self.accept()
